#!/usr/bin/env python3


class NoSuchTagError(Exception):
    def __init__(self, tag):
        self.tag = tag


class AmbiguousTagError(Exception):
    def __init__(self, tag, endpoints):
        self.tag = tag
        self.endpoints = endpoints


class RequestError(Exception):
    def __init__(self, url, status, body):
        self.url = url
        self.status = status
        self.body = body



#!/usr/bin/env python

import argparse
import sys
from os import getenv

from .client import Portainer
from .errors import RequestError


def parse_mount(conf: str):
    try:
        path, name = conf.split(":")
        return (path, name)
    except Exception:
        print("invalid mount argument: " + conf)


def read_string(path):
    with open(path, "r") as f:
        return "".join(f.readlines())


def deploy_app(args):
    try:
        deploy(args)
    except RequestError as err:
        print(err.url, err.status, err.body, file=sys.stderr)
        sys.exit("Failed to deploy!")


def deploy(args):
    HOST = getenv("PORTAINER_HOST", args.host)
    PASSWORD = getenv("PORTAINER_PASSWORD", args.password)
    USERNAME = getenv("PORTAINER_USERNAME", args.username)
    API_TOKEN = getenv("PORTAINER_TOKEN", args.api_token)

    bot = Portainer(HOST)
    if API_TOKEN:
        bot.authorize(API_TOKEN)
    else:
        bot.login(USERNAME, PASSWORD)

    endpoint = bot.get_endpoint_by_tag(args.environment)

    compose = ""
    with open(args.compose_file, "r") as f:
        compose = "".join(f.readlines())

    env_lines = args.variable
    if args.env_file != None:
        with open(args.env_file, "r") as f:
            lines = f.readlines()
            env_lines = env_lines + lines

    env_vars = []
    for line in env_lines:
        striped = line.strip()
        idx = striped.find("=")
        env_vars.append({"name": striped[:idx], "value": striped[idx + 1 :]})

    for conf in args.config:
        path, name = parse_mount(conf)
        data = read_string(path)
        endpoint.create_config(name=name, data=data)

    for conf in args.secret:
        path, name = parse_mount(conf)
        data = read_string(path)
        endpoint.create_secret(name=name, data=data)

    stack_name = (
        args.stack_name
        if args.stack_name != None
        else args.environment + "-" + args.name
    )

    endpoint.deploy(stack_name=stack_name, compose=compose, env_vars=env_vars)


def destroy(args):
    print(args)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Portainer deployment client",
        epilog="Use it to automate workflows for less mouse clicks!",
    )
    ### HACK: this is due to a known issue in argparse in python3
    parser.set_defaults(func=lambda args: parser.print_help())

    parser.add_argument(
        "-T",
        "--api-token",
        help="api token for user, overrides PORTAINER_TOKEN variable",
    )
    parser.add_argument(
        "-H",
        "--host",
        help="portainer host, overrides PORTAINER_HOST variable; defaults to `http://localhost`",
        default="http://localhost",
    )
    parser.add_argument(
        "-U",
        "--username",
        help="username to login, overrides PORTAINER_USERNAME variable; defaults to `admin`",
        default="admin",
    )
    parser.add_argument(
        "-P",
        "--password",
        help="password for user, overrides PORTAINER_PASSWORD variable; defaults to admin",
        default="admin",
    )

    subparsers = parser.add_subparsers(
        title="subcommands", description="valid subcommands", help="additional help"
    )
    deploy_cmd = subparsers.add_parser("deploy")
    deploy_cmd.add_argument(
        "-f",
        "--compose-file",
        help="compose manifest file",
        default="docker-compose.yaml",
        required=True,
    )
    deploy_cmd.add_argument("-n", "--name", help="deployment name", required=True)
    deploy_cmd.add_argument(
        "-E",
        "--environment",
        help="environment to deploy on",
        choices=["staging", "production"],
        required=True,
    )
    deploy_cmd.add_argument(
        "-S", "--stack-name", help="use this to override stack name"
    )
    deploy_cmd.add_argument(
        "--env-file",
        help="dot env file used for deployment, it will be used as stack environment in portainer",
    )
    deploy_cmd.add_argument(
        "-e",
        "--variable",
        action="append",
        help="environment variable `SOME_ENV=some-value`",
        default=[],
    )
    deploy_cmd.add_argument(
        "-c",
        "--config",
        help="""create config; args must be like `local-path-to-file:conf-name`;
  NOTE that as configs are immutable and might be already in use, your config name must not exist!
  use versioning or date in names to always get a new name
  """,
        action="append",
        default=[],
    )
    deploy_cmd.add_argument(
        "-s",
        "--secret",
        help="create a new secret; see --config.",
        action="append",
        default=[],
    )
    deploy_cmd.set_defaults(func=deploy_app)

    destroy_cmd = subparsers.add_parser("destroy")
    destroy_cmd.add_argument("-n", "--name", help="deployment name", required=True)
    destroy_cmd.set_defaults(func=destroy)

    return parser


def main():
    parser = build_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import json
import logging

import requests
from requests.structures import CaseInsensitiveDict

from . import errors, helpers

logging.basicConfig()
logger = logging.getLogger("portainer")
logger.setLevel(logging.DEBUG)


class Portainer:
    def __init__(self, host: str):
        self.host = host
        self.token: CaseInsensitiveDict = CaseInsensitiveDict()

    def __extract(self, resp):
        if resp.ok:
            return resp.json()
        else:
            raise errors.RequestError(resp.url, resp.status_code, resp.text)

    def get(self, url: str):
        resp = requests.get(self.host + url, headers=self.token)
        return self.__extract(resp)

    def post(self, url: str, data):
        resp = requests.post(self.host + url, headers=self.token, data=json.dumps(data))
        return self.__extract(resp)

    def put(self, url: str, data):
        resp = requests.put(self.host + url, headers=self.token, data=json.dumps(data))
        return self.__extract(resp)

    def delete(self, url: str):
        resp = requests.post(self.host + url, headers=self.token)
        return self.__extract(resp)

    def authorize(self, api_token: str):
        logger.info("Authorized for " + self.host + " using api token")
        self.token["Accept"] = "application/json"
        self.token["x-api-key"] = api_token
        return

    def login(self, username: str, password: str):
        logger.info("Trying to login to " + self.host + "...")
        body = {"username": username, "password": password}
        resp = self.post("/auth", body)
        token = resp["jwt"]
        self.token["Accept"] = "application/json"
        self.token["Authorization"] = "Bearer " + token
        return

    def list_stacks(self):
        logger.info("Getting a list of stacks")
        resp = self.get("/stacks")
        return resp

    def get_stack_id(self, name):
        stacks = self.list_stacks()
        ids = [s["Id"] for s in stacks if s["Name"] == name]
        return ids[0]

    def get_stack(self, name):
        stacks = self.list_stacks()
        filtered = [s for s in stacks if s["Name"] == name]
        return filtered

    def list_endpoints(self):
        logger.info("Getting a list of endpoints")
        resp = self.get("/endpoints")
        return resp

    def list_tags(self):
        logger.info("Getting a list of tags")
        resp = self.get("/tags")
        return resp

    def get_endpoint_id(self, name):
        endpoints = self.list_endpoints()
        ids = [e["Id"] for e in endpoints if e["Name"] == name]
        return ids[0]

    def get_endpoint_by_id(self, e_id):
        return Endpoint(client=self, endpoint_id=e_id)

    def get_endpoint_by_tag(self, tag):
        tags = self.list_tags()
        target = [ei for t in tags for ei in t["Endpoints"] if t["Name"] == tag]
        if len(target) == 1:
            return self.get_endpoint_by_id(target[0])
        elif len(target) == 0:
            raise errors.NoSuchTagError(tag)
        else:
            raise errors.AmbiguousTagError(tag, target)


class Endpoint:
    def __init__(self, client, endpoint_id):
        self.client = client
        self.endpoint_id = str(endpoint_id)

    def get_docker_info(self):
        logger.info("Getting endpoint info")
        resp = self.client.get("/endpoints/" + self.endpoint_id + "/docker/info")
        return resp

    def create_config(self, name, data):
        name = name.strip()

        logger.info("Creating new config " + name)
        body = {"Data": helpers.to_base64(data), "Name": name, "Labels": {}}
        try:
            self.client.post(
                "/endpoints/" + self.endpoint_id + "/docker/configs/create", body
            )
        except Exception as e:
            logger.error(f"cannot create config: {str(e)}")

    def create_secret(self, name, data):
        name = name.strip()

        logger.info("Creating new secret " + name)
        body = {"Data": helpers.to_base64(data), "Name": name, "Labels": {}}
        try:
            self.client.post(
                "/endpoints/" + self.endpoint_id + "/docker/secrets/create", body
            )
        except Exception as e:
            logger.error(f"cannot create secret: {str(e)}")

    def deploy(self, stack_name: str, compose: str, env_vars):
        stack_name = stack_name.lower().strip()

        stacks = self.client.get_stack(stack_name)

        if len(stacks) == 1:
            logger.info("Updating existing stack name: " + stack_name)
            existing_stack = stacks[0]
            stack_id = existing_stack["Id"]

            data = {
                "Prune": True,
                "StackFileContent": compose,
                "Env": env_vars,
                "id": stack_id,
            }

            self.client.put(
                "/stacks/" + str(stack_id) + "?endpointId=" + self.endpoint_id, data
            )

        else:
            logger.info("No existing stack with name: " + stack_name)
            info = self.get_docker_info()
            swarm_id = info["Swarm"]["Cluster"]["ID"]

            data = {
                "Env": env_vars,
                "Name": stack_name,
                "SwarmID": swarm_id,
                "StackFileContent": compose,
            }
            self.client.post(
                "/stacks?endpointId=" + self.endpoint_id + "&method=string&type=1",
                data=data,
            )

        return
